/*
**++
**  FACILITY:
**      NEWSDEFINE.H
**
**  ABSTRACT:
**      Common NEWS constant definitions.
**
**  AUTHOR:
**      Geoff Huston
**
**  COPYRIGHT:
**      Copyright  1988,1989,1990,1991
**--
**/

#ifndef _NEWSDEFINE_H

#define _NEWSDEFINE_H

#include "newssite.h"
#include <stdio.h>

#ifdef	__GNUC__
		/* GNU CC settings */
/*
 * Can't use VAXC$ESTABLISH() with v1.x (actually 2.2 or higher is needed
 * to use it, but there's no way to test the minor version number).
 * If using 2.0 or 2.1, then compile with /define="GNUC_HANDLER_HACK".
 */
#if __GNUC__ < 2
#define GNUC_HANDLER_HACK 1
#define C$_LONGJMP 0x0035C05C	/* from module C$ERRNO in VAXCRTL */
#endif
#endif

#ifndef GNUC_HANDLER_HACK
#define GNUC_HANDLER_HACK 0
#endif

/*
 *  PROTOTYPES is something that says whether we can define functions with
 *  argument prototypes, as allowed by ANSI C.
 */

#if vax11c
/* note: viable for VAX C v2.3 and later; not for v2.2 or earlier */
#define PROTOTYPES	1
#endif

#ifndef	PROTOTYPES
#ifdef	__STDC__
#define	PROTOTYPES	1
#else
#define	PROTOTYPES	0
#endif
#endif

#define MAXQUALLEN      132

#define IO_SIZE         512             /* various string sizes */
#define FILE_SIZE       256
#define NODE_SIZE       256
#define NGRP_SIZE       256

#define _c$cks(s)	(((status = (s)) & 1) ? 1 : (lib$signal(status),0))
#ifdef vax11c
#define c$ac(v)		&(v)
#else
#define c$ac(v)		c$rfi(v)
#endif

#define max(a,b)        (((a) > (b)) ? (a) : (b))
#define min(a,b)        (((a) > (b)) ? (b) : (a))
#define	empty_string(s)	(*(s) == '\0')

#define DAY_SECS        86400

#define HLP$V_PROMPT    0

#define DIR_ALL         0
#define DIR_DATE        1
#define DIR_NEW         2
#define DIR_REGISTER    3
#define DIR_UNREGISTER  4

#define HELP_FILE       "NEWS_ROOT:NEWS.HLB"

#define ALIAS_FILE      "NEWS_MANAGER:NEWS.ALIASES"
#define ADDRESS_FILE    "NEWS_MANAGER:NEWS_ADDRESS.CNF"
#define MAILPATHS_FILE  "NEWS_MANAGER:MAILPATHS"
#define SYS_FILE        "NEWS_MANAGER:NEWS.SYS"
#define SYSD_FILE       "NEWS_MANAGER:NEWS.SYSD"
#define DIST_FILE       "NEWS_MANAGER:NEWS.DISTRIBUTION"
#define DISTRIBUTIONS_FILE	"NEWS_MANAGER:NEWS.DISTRIBUTIONS"
#define POSTCC_FILE     "NEWS_MANAGER:NEWS_POST.CC"

#define NNTP_ACCESS_FILE    "NEWS_ROOT:NNTP_ACCESS.NEWS"

#define HIST_FILE       "NEWS_ROOT:HISTORY.V60"

#define HIST_OFF       "NEWS_ROOT:HISTORY.OFF"
#define HIST_VAL        45

#define User_file       "SYS$LOGIN:NEWSRC"
#define Itm_files       "NEWS_DEVICE:[000000...]*.ITM;*"
#define Itm_template    "NEWS_DEVICE:[%s]%d.ITM"
#define Itm_device	"NEWS_DEVICE:"
#define Itmg_template   "NEWS_DEVICE:[%s]*.ITM;*"
#define Grp_template    "NEWS_DEVICE:[%s]"
#define Dir_template    "NEWS_DEVICE:[000000.%s^DIR"
#define Access_template "NEWS_DEVICE:[%s]ACCESS.NEWS"
#define News_ini        "NEWS_INI"

#define DISP_BUF_SIZE   2000

#define IDLEN           60
#define SUBJLEN 	50                  /* max length of newsgroup name */
#define FROMLEN		60

/*
 * newsitem file: record structure
 */

#define ITM_FILENAME   "NEWS_ROOT:NEWS.ITEMS"

typedef struct {
    unsigned int itm_num;           /* newsitem index [key 0, b 1-4] 4   4 */
    unsigned int itm_grp;           /* newsgroup index[key 0, b 5-8] 4   8 */
                                    /*                [key 1, seg 1]       */
    char itm_id[IDLEN];             /* item id        [key 1, seg 0]60  68 */
    unsigned int itm_recvdate;      /* recvdate                      4  72 */
    unsigned int itm_lines;         /* lines                         4  76 */
    char itm_title[SUBJLEN];        /* title                        50 126 */
    unsigned int itm_cachedate;     /* server - cache entry date     4 130 */
    unsigned short itm_life;        /* retention period              2 132 */
    unsigned int itm_flags;         /* item flags                    4 136 */
    unsigned int itm_cid;           /* conversation id [key 2, b1-4] 4 140 */
    unsigned int itm_postdate;      /* posted date                   4 144 */
    char itm_from[FROMLEN];         /* poster                       60 204 */  
    } ITM, *ITM_PTR;

/*
 * itm_flags - bit definitions
 */

#define NEWS_M_NEW              1       /* new item */
#define NEWS_M_UNREAD           2       /* unread item */
#define NEWS_M_XPOST		4	/* local xposting to another newsgroup*/
#define NEWS_M_MAILITEM		8
#define NEWS_M_MAILMARK		16
#define NEWS_M_MAILREPLIED	32
#define NEWS_M_NOACCESS		64
#define NEWS_M_LINESVALID	128

#define NEWS_V_NEW              0
#define NEWS_V_UNREAD           1
#define NEWS_V_XPOST		2
#define NEWS_V_MAILITEM		3
#define NEWS_V_MAILMARK		4
#define NEWS_V_MAILREPLIED	5
#define NEWS_V_NOACCESS		6
#define NEWS_V_LINESVALID	7

/*
 * newsgroup file: record structure
 */

#define GRP_FILENAME   "NEWS_ROOT:NEWS.GROUPS"

typedef struct {
    char grp_name[SUBJLEN];         /* newsgroup name  [key 0]      50  50 */
    unsigned int grp_num;           /* newsgroup index [key 1]       4  54 */
    unsigned int grp_topnum;        /* highest item number used      4  58 */
    unsigned int grp_count;         /* item count                    4  62 */
    unsigned int grp_credate;       /* date created (unixtime)       4  66 */
    unsigned int grp_entdate;       /* last entry date (unixtime)    4  70 */
    unsigned short grp_life;        /* newsgroup hold time (days)    2  72 */
    unsigned short grp_itmlife;     /* item retention time (days)    2  74 */
    unsigned short grp_srvcache;    /* NNTP client cache time        2  76 */
    unsigned int grp_flags;         /* newsgroup flags               4  80 */
    char grp_srvproto;              /* NNTP transport protocol       1  81 */
    char grp_srvnode[132];          /* NNTP server node            132 213 */
    char grp_topic[125];            /* group topic line            125 338 */
    char grp_notice[125];           /* group notice line           125 463 */
                                    /* end of file area - remainder of the
                                       record is only used when loaded     */
    int grp_reg;                    /* registered level              4 467 */
    unsigned int grp_unread;        /* unread count                  4 471 */
    ITM_PTR grp_ia;                 /* item array                    4 475 */
    unsigned int grp_iasize;        /* item array size               4 479 */
    int grp_iavd;                   /* item virtual display          4 483 */
    int grp_iavdsize;               /* item virtual disp size        4 487 */
    int grp_iapaste;                /* item display paste pos        4 491 */
    unsigned int grp_c_itm;         /* item current index            4 495 */
    int grp_display_indx;           /* group display position        4 499 */
    char *grp_reg_text;             /* register test pointer         4 503 */
    unsigned int grp_topreadnum;    /* top read number               4 507 */
    unsigned int grp_firstnum;	    /* lowest article number         4 511 */
    unsigned char grp_displaystk;   /* used by unseenstack	     1 512 */
    } GRP, *GRP_PTR;

/*
 * grp_flags - bit definitions
 */

#define NEWS_M_LOCAL            1       /* non-network newsgroup */
#define NEWS_M_WRITE_ACCESS     2       /* user post access */
#define NEWS_M_MOD_ACCESS       4       /* user moderator access enabled */
#define NEWS_M_MOD_USER         8       /* user moderator */
#define NEWS_M_NOWRITE_SET      16      /* newsgroup is set nowrite */
#define NEWS_M_RESTRICT_SET     32      /* newsgroup is set restricted */
#define NEWS_M_ACCESS_CHECKED   64      /* access file checked ? */
#define NEWS_M_NNTPSRV          128     /* NNTP server defined */
#define NEWS_M_NNTPCACHE        256     /* NNTP server cache enabled */
#define NEWS_M_MAILMODERATE     512     /* newsgroup is moderated */
#define NEWS_M_OLDGROUP        1024     /* newsgroup is moderated */
#define NEWS_M_NOAPPROVAL      2048	/* moderated newsgroup - no approval */
#define NEWS_M_MAILLIST	       4096	/* this is a mailing list */
#define NEWS_M_MAILGROUP       8192	/* this is a vms mail folder */
#define NEWS_M_NEWSKIP	      16384
#define NEWS_M_IGNEXP	      32768	/* ignore expiry headers */
#define NEWS_M_NOSTORE	      65536	/* dont store item texts locally */
#define NEWS_M_UNSEENSTACK   131072
#define NEWS_M_UNSEENITEMS   262144
#define NEWS_M_NONDEFDISPLAY 524288
#define NEWS_M_ORDERBYSUBJECT 1048576
#define NEWS_M_READTHISSESSION 2097152
#define NEWS_M_NNTPSRVBYNUM    4194304

#define NEWS_V_LOCAL            0
#define NEWS_V_WRITE_ACCESS     1
#define NEWS_V_MOD_ACCESS       2
#define NEWS_V_MOD_USER         3
#define NEWS_V_NOWRITE_SET      4
#define NEWS_V_RESTRICT_SET     5
#define NEWS_V_ACCESS_CHECKED   6
#define NEWS_V_NNTPSRV          7
#define NEWS_V_NNTPCACHE        8
#define NEWS_V_MAILMODERATE     9
#define NEWS_V_OLDGROUP        10
#define NEWS_V_NOAPPROVAL      11
#define NEWS_V_MAILLIST	       12
#define NEWS_V_MAILGROUP       13
#define NEWS_V_NEWSKIP	       14
#define NEWS_V_IGNEXP		15
#define NEWS_V_NOSTORE		16
#define NEWS_V_UNSEENSTACK	17
#define NEWS_V_UNSEENITEMS	18
#define NEWS_V_NONDEFDISPLAY    19
#define NEWS_V_ORDERBYSUBJECT   20
#define NEWS_V_READTHISSESSION  21
#define NEWS_V_NNTPSRVBYNUM     22

struct itms {
        char itmid[IDLEN];
        int itmdate,
            itmsize;
        char itmtitle[SUBJLEN];
        char itmfrom[FROMLEN];
        int itmvalid;
        struct itms *itmsnext;
        };

struct grps {
        char grpname[SUBJLEN];
        char grpnode[132];
        int  nproto;
        unsigned int itmgrp;
        int grptime;
        struct itms *ids,
                    *idr;
        struct grps *grpsnext;
        };

typedef struct its {
    int ngrp;
    int fitm, litm;
    } NOTE_RANGE;

typedef struct dptt {
  int do_header,
      usr_line,
      gets_size,
      text_lines,
      text_cline,
      text_displayed,
      disp_size,
      shdid,
      sdid,
      sdid_paste,
      sdid_home,
      sdid_line,
      sdid_baseline,
      sdid_dsize,
      sdid_current,
      sheader_line,
      target_paste,
      fdepth,
      str_found,
      str_searching,
      str_case_sensitive,
      str_length,
      str_line,
      str_col,
      display_code;
  char header_line[500],
       str_target[512];
  struct stat dispf_stat;
  } DPT, *DPTR;

/* From NEWSFILES.C -	Tim Cook, 12 Feb 1990 */
struct rms_file {
  struct FAB f;
  struct RAB r;
  struct XABPRO p;
  };

/* From NEWSRC.C -	Tim Cook, 12 Feb 1990 */
struct marks {
    int m_type;
    char *m_tag;
    unsigned int *m_grps;
    unsigned int *m_itms;
    int m_size;
    int m_malloc;
    char *k_grp_name;
    char *k_itm_title;
    char *k_itm_from;
    char *k_itm_header;
    int k_test_from;
    int k_test_header;
    int k_test;
    struct marks *m_next;
    };

#define OHEADERS 18
#define NHEADERS 21

#define PATH          0
#define NEWSGROUPS    1
#define SUBJECT       2
#define MESSAGE_ID    3
#define FROM          4
#define DATE          5
#define REPLY_TO      6
#define SENDER        7
#define FOLLOWUP_TO   8
#define EXPIRES       9
#define REFERENCES    10
#define CONTROL       11
#define DISTRIBUTION  12
#define ORGANIZATION  13
#define KEYWORDS      14
#define SUMMARY       15
#define APPROVED      16
#define SUPERSEDES    17
#define LINES         18
#define XREF          19
#define RELAY_VERSION 20

#ifndef NNTP_CLIENT
#define NNTP_CLIENT	1
#endif

#ifndef NNTP_CLIENT_ONLY
#define NNTP_CLIENT_ONLY	0
#endif

#if NNTP_CLIENT_ONLY
#define NNTP_CLIENT	1
#endif

#define OUTER_LOOP	0
#define I_DISPLAY_LOOP	1
#define T_DISPLAY_LOOP	2

#define MAXTARGETS	10		/* maximum nuymber of ANDed search target strings */
#define MAXIDLEN	1023		/* maximum size of a message-id string */

typedef struct IOSB_TYPE {		/* I/O status block type */
		 unsigned short	status;
		 unsigned short	count;
		 unsigned short	terminator;
		 unsigned short	term_count;
		 } IOSB_DEF;

typedef struct dir_class {		/* class structure */
    char *c_name;
    unsigned int *c_nums;
    int c_size;
    int c_malloc;
    struct dir_class *c_next;
    } DIR_CLASS, *DIR_CLASS_PTR;

typedef struct als {
  char *ngname;
  char *alias_name;
  struct als *a_next;
  } ALIAS, *ALIAS_PTR;

typedef struct sys_filter {
    char *sys_filtnam;
    struct sys_filter *sys_fnext;
    } SYS_FILTER_T, NAME_LIST;

typedef struct sys_entry {
    char *sys_nodename;
    SYS_FILTER_T *sys_no;
    SYS_FILTER_T *sys_f;
    SYS_FILTER_T *sys_d;
    char *sys_type;
    char *sys_file;
    int sys_batch_size;
    int sys_article_count;
    int sys_article_bytes;
    int sys_id_count;		/* IHAVE ID's */
    FILE *sys_fp;
    stat_t sys_statb;
    struct sys_entry *sys_enext;
    } SYS_ENTRY_T;

typedef struct dist_entry {
    char *from_name;
    SYS_FILTER_T *to_names;
    struct dist_entry *next;
    } DIST_ENTRY_T;
				/* Add definition of TCP modules */
#ifdef CMU
#define CMU	1
#endif

#ifdef TWG
#define TWG	1
#endif

#ifdef MULTINET
#define MULTINET	1
#define __TYPES_OFF_T
#define __TYPES_DEV_T
#endif

/*----- NEWSADD/NEWSCONTROL Control: message parse routines exit status values */

#define CTLOK			1
#define CTLOKNOCHG		3

#define NOSCRATCHFILE	 	2
#define NORETURNADDRESS	 	4
#define NEWGROUPERROR	 	6
#define NONEWSGROUPFIELD 	8
#define NONAPPROVEDCONTROL	10
#define LOCALNEWSGROUP		12
#define NORMNEWSGROUP		14
#define NOACCESSINFILE		16
#define NOSENDERADDRESS		18
#define BADFORMAT		20
#define NOTTOME			22
#define NOREMOTESYS		24
#define UNKCTL			26

/*----- NEWSADD/NEWSCONTROL Control: switch flag definitions */
#define EXECUTE_CONTROL		1
#define EXECUTE_HOLD		1
#define EXECUTE_DELETE		2
#define EXECUTE_LOCAL		4

/*----- NEWSADD Parser output codes */

#define SCAN1OK			1

#define NOPATH		        2
#define NOGROUPS		4
#define NOSUBJ			8
#define NOID			16
#define NOFROM			32
#define NODATE			64
#define BADID			128

#define	NOFILENAME		 2
#define NOREADACCESS		 4
#define NOSCRFILEACCESS		 6
#define BADHD_STARTSPACE	 8
#define BADHD_NOCOLON		10
#define BADHD_NOTAKEYWORD	12
#define BADHD_REPEATEDKEYWORD	14
#define EMPTYINFILE		16
#define NOLOCALNEWSGROUPS	18
#define JUNKED			20
#define KNOWNID			22
#define NOTLOCALLYACCEPTED	24

/*--------- NEWSPROFILE flag settings */

#define PROFILE_AUTO       	1
#define PROFILE_CLASS       	2
#define PROFILE_EDIT        	4
#define PROFILE_PERSNAME    	8
#define PROFILE_PRINT       	16
#define PROFILE_DIRECTORY   	32
#define PROFILE_NEWREGISTER 	64
#define PROFILE_SCANSIZE	128
#define PROFILE_SIGNATURE	256
#define PROFILE_LOCSIGNATURE	512
#define PROFILE_ORGANIZATION	1024
#define PROFILE_KEYDEFS		2048
#define PROFILE_SEARCH		4096
#define PROFILE_DISPLAY		8192
#define PROFILE_LINEEDIT	16384
#define PROFILE_BROADCAST	32768
#define PROFILE_UNSEENSTACK	65536		/* bit 16 */
#define PROFILE_UNSEENITEMS	131072
#define PROFILE_RCREG		262144
#define PROFILE_SHOWDIRS	524288
#define PROFILE_FASTLOAD	1048576
#define PROFILE_REPLYTO		2097152
#define PROFILE_VIEW            4194304

#define PROFILE_UNSEEN      1
#define PROFILE_DIR         2

#define SHOW_CLASSES		1
#define SHOW_ALIASES		2
#define DIR_ITEMS		3
#define DIR_GROUPS		4
#define SHOW_SYS		5
#define SHOW_ITEM		6
#define SHOW_FILE		7
#define SHOW_NOTE		8
#define SHOW_MAILPROFILE	9
#define SHOW_PROFILE		10
#define SHOW_MESSAGES		11
#define SHOW_NEWSGROUP		12
#define SHOW_SUPERVISOR		13
#define SHOW_MODERATOR		14
#define SHOW_MEMBER		15
#define SHOW_ENTRY		16
#define SHOW_MARK		17

#endif	/* _NEWSDEFINE_H */
