/*
**++
**  FACILITY:
**	NNTP_XMIT
**
**  ABSTRACT:
**	Use the NNTP IHAVE command to manage NEWS distribution
**
**  AUTHOR:
**	Geoff Huston
**
**  COPYRIGHT:
**	Copyright  1988,1989,1990,1991,1992
**
**  CONTRIBUTIONS:
**	UCX Support botched in by D.J.Young (CNBR10@UK.AC.STRATH.VAXA)
**	16th August 1990
**	Numerous bug fixes & improvements: Bill Glass glass@vixvax.mgi.com
**	Report on xfer rates: Earle Ake ake@dayton.saic.com
**
**  VERSION:
**	V6.0-4	25-May-1991	volz@process.com
**	  - Added TCPware support
**	V6.1	 8-Feb-1992	Mark Wood - imhw400@indyvax.iupui.edu
**	  - Bug fix to command line parse
**	V6.1	 9-Feb-1992	David E. Bellamy - bellamy@commerce.uq.oz.au
**	  - Bug fix to UCX socket port assignments
**      V6.1     9-Feb-1992	kjb@calmasd.prime.com
**        - Modified win/tcp support to include one more level of indirection
**          to access destination host address
**      V6.1     21-Feb-1992	sloane@kuhub.cc.ukans.edu
**        - Fix memory leak
**
**--
*/

#module NNTP_XMIT "V6.1"

void close_net();

#include "newsinclude.h"
#include "newsdefine.h"

#include iodef
#include lnmdef
#include psldef
#include signal

#define LJP		1
#define LOG_TABLE	"LNM$SYSTEM"

#define	 DECNET		0
#define	 CMUTCP		1
#define	 WINTCP		2
#define	 MULTINETTCP	3
#define	 UCXTCP		4
#define  TCPWARETCP	6

#ifdef CMU
#define CMU	1
#endif

#ifdef TWG
#define	 TWG	    1
#endif

#ifdef MULTINET
#define	 MULTINET	 1
#define __TYPES_OFF_T
#define __TYPES_DEV_T
#endif

#ifdef TCPWARE
#define  TCPWARE	1
#define __TYPES_H_DEFINED
#endif

#if TWG || MULTINET || TCPWARE
#include <types.h>
#include <socket.h>
#include <netdb.h>
#include <in.h>
#include <inetiodef.h>
#define NNTP_PORT	(119)
char *addr_pointer;             /*kjb@calmasd.prime.com*/
struct hostent *dest_host;
struct sockaddr_in data_socket = {0};
#endif

#ifdef UCX
#define UCX     1
#define htons(x) ((unsigned short)((x<<8)|(x>>8)))
#define NNTP_PORT    (119)
#endif

#if UCX
#include <ucx$inetdef.h>
struct sockaddr {
  short inet_family;
  short inet_port;
  int inet_adrs;
  char bklb[8];
  };

struct itlist { int lgth; struct sockaddr *hst; };

static short sck_parm[2];
static struct sockaddr local_host, remote_host;
struct itlist lhst_adrs, rhst_adrs;
#endif

					/* compile time constants */
#define CLIENT_TIMER	250
#define RESP_TIMER	30
#define NNTP_MAX_OFFER	5000

					/* program constants */
#define X_BUF_SIZE	1024
#define NNTP_STRLEN	1024
#define OUT_BUF_TERM	"\n"

#define ERROR		0
#define MAX_OFFERED	1
#define MANAGER_STOP	2


					/* premature exit - clean up as far as possible */
#define GO_DIE(v) {					\
		    if (fpw) {				\
		      fprintf(fpw,"%s\n",inpline);	\
		      ++lines_written;			\
		      }					\
		    premature_break((v),fpr,fpi,fpw,logging,lines_written);\
		  }

globalref int nntpxmitcmd;

			/* provide a fake definition of newscmd and the CLI tables are
		           not required, but are referenced in newsvariables.h */
#if vax11c
globaldef int newscmd;
#else
void newscmd() {}
#endif

					/* program variables */
char net_open_chan[128] = {""},
     ibuf[X_BUF_SIZE],
     obuf[512],
     xbuf[X_BUF_SIZE + 1],
     client_msg[512],
     *rptr,
     lhost[128],
     node[128],
     datum[32],
     logfile[256],
     dbgfile[256],
     idfile[256],
     tmpf[256],
     taskstr[132];

unsigned short net_chan,
	       trm;

int net_read_status = 0,
    hiber_state = 0,
    cmd_code,
    net_proto = 0,
    debugging = 0,
    decnetstream = 0,
    status,
    *c$_tmphead = 0,	     /* head of side list */
    failed = 0,
    expired = 0,
    accepted = 0,
    rejected = 0,
    offered = 0,
    idx = 0;

long bytes_sent = 0,
     total_bytes_sent = 0;
int  total_elapsed_time = 0;

FILE *fpd;

struct iosb {
	unsigned short iostatus;
	unsigned short iosize;
	int netinfo;
	} read_iosb,
	  write_iosb;

struct FAB grpfab,		/* newsgroup file fab */
	   itmfab;		/* newsitem file fab */
struct RAB grprab,		/* newsgroup file rab */
	   itmrab;		/* newsitem file rab */

ITM newsitm;			/* newsitem i/o buffer */
GRP newsgrp;			/* newsgroup i/o buffer */

closefiles()
{
  return(1);
}

/*
 *  idx_file routines - RMS indexed identifier file support
 */

struct FAB  idxfab;
struct RAB  idxrab;
struct XABKEY  idxkey_1;
struct XABPRO  idxpro_1;

struct idx  {char idx_id[MAXIDLEN+1];}	newsidx;

/*
 *  open_id_file
 *
 *  Open an identifier file - create the file if it does not exist already.
 */

open_id_file(idx_name)	char  *idx_name;
{
  idxfab = cc$rms_fab;
  idxfab.fab$b_fac = FAB$M_GET | FAB$M_DEL;
  idxfab.fab$l_fna = idx_name;
  idxfab.fab$b_fns = strlen(idxfab.fab$l_fna);
  idxfab.fab$l_dna = "NEWS_MANAGER:IHAVE.IDX";
  idxfab.fab$b_dns = strlen(idxfab.fab$l_dna);
  idxfab.fab$b_shr = FAB$M_SHRDEL | FAB$M_SHRGET | FAB$M_SHRPUT | FAB$M_SHRUPD;

  sysprv();
  if (!(sys$open(&idxfab) & 1)) return(nosysprv(),0);

  idxrab = cc$rms_rab;
  idxrab.rab$b_krf = 0;
  idxrab.rab$l_fab = &idxfab;
  idxrab.rab$l_ubf = &newsidx;
  idxrab.rab$w_usz = sizeof newsidx;
  idxrab.rab$l_rbf = &newsidx;
  idxrab.rab$w_rsz = sizeof newsidx;

  if (!(sys$connect(&idxrab) & 1)) {
    sys$close(&idxfab);
    nosysprv();
    return(0);
    }
  nosysprv();
  idxrab.rab$b_rac = RAB$C_SEQ;
  idxrab.rab$b_krf = 0;
  sys$rewind(&idxrab);
  return(1);
}

/*
 * close_id_file
 *
 * close the index file - use sysprv - cause that was used to open it
 */

close_id_file()
{
  sysprv();
  sys$close(&idxfab);
  nosysprv();
}

/*
 * read_id
 *
 * Get the next entry from the ID file
 */

read_id(id)	char  *id;
{
  if (sys$get(&idxrab) & 1) {
    strcpy(id,newsidx.idx_id);
    return(1);
    }
  return(*id = '\0');
}

/*
 * delete_id
 *
 * Delete the entry from the ID file
 */

delete_id()
{
  sys$delete(&idxrab);
}

/*
 * dat
 *
 * Convert the time into a timestring
 */

dat()
{
  struct tm *stm;
  int cur_time;
  char *p;

  time(&cur_time);
  p = ctime(&cur_time);
  p += 4;
  stm = localtime(&cur_time);
  sprintf(datum,"%d %.3s %d %02d:%02d:%02d ",
	stm->tm_mday,p,stm->tm_year,stm->tm_hour,stm->tm_min,stm->tm_sec);
}

/*
 * elapsed_time
 *
 * get the elapsed time in hundredths (0.01) of seconds
 *
 * usage:
 *
 *	elasped_time(1); will perform a 'start-the-clock'
 *
 *	elapsed_time(0); returns the number of centi-seconds since the last
 *		'start-the-clock' command
 */

#include timeb

static int elapsed_time(update)
  int update;
{
  int result;
  static struct timeb rbefore, rafter;

  ftime(&rafter);
  result = 100 * (rafter.time - rbefore.time) +
	((int) rafter.millitm - (int) rbefore.millitm) / 10;
  if (update) rbefore = rafter;
  return(result);
}

/*
 * manager_say_stop
 *
 * Check for the /sys/exec logical name NEWS_STOP being defined
 */

manager_say_stop()
{
  int len;
  char exec_mode = PSL$C_EXEC,
       translate_value[256];
  unsigned int trnlist[4]={(LNM$_STRING << 16) + LNM$C_NAMLENGTH,
			     (&translate_value), (&len), 0};

  return ((sys$trnlnm(0,c$dsc(LOG_TABLE),c$dsc("NEWS_STOP"),&exec_mode,&trnlist) & 1));
}

/*  Stack allocation routines
 *
 *  c$alloc_tmp
 *
 *  Allocate temporary storage for use with c$ calls
 */

int *c$alloc_tmp(size)	int size;
{
  int *tmp;

  tmp = (int *) malloc(size + 4);
  *tmp = c$_tmphead;
  c$_tmphead = tmp;
  return(c$_tmphead + 1);
}

/*
 *  c$dsc
 *
 * Allocate a string descriptor in the temp arg area
 */

c$dsc(str)	char *str;
{
  struct dsc$descriptor *tmpdsc;

  tmpdsc = c$alloc_tmp(8);
  tmpdsc->dsc$w_length = strlen(str);
  tmpdsc->dsc$b_dtype = DSC$K_DTYPE_T;
  tmpdsc->dsc$b_class = DSC$K_CLASS_S;
  tmpdsc->dsc$a_pointer = str;
  return(tmpdsc);
}

/*
 *  c$cks
 *
 *  Check return status, signal errors, and free the temp argument area
 */

c$cks(status)	int status;
{
  int *tmp;

  if (!(status & 1)) lib$signal(status);
  while (c$_tmphead) {
    tmp = *c$_tmphead;
    free(c$_tmphead);
    c$_tmphead = tmp;
    }
  return(status);
}

/*
 *  nosysprv
 *
 *  Turn off installed sysprv - BUT if user already had sysprv from AUTHORIZE
 *  then leave it on (as there is no change in user functionality)
 */

nosysprv()
{
  unsigned int authprivs[2], msysprv[2] ;
  int item = JPI$_PROCPRIV;

  msysprv[0] = PRV$M_SYSPRV;
  msysprv[1] = 0;
  c$cks(lib$getjpi(&item,0,0,&authprivs,0,0));
  if (!(authprivs[0] & PRV$M_SYSPRV)) c$cks(sys$setprv(0,msysprv,0,0));
}

/*
 *  sysprv
 *
 *  Turn sysprv on as a temp priv - assumes image was installed with SYSPRV
 */

sysprv()
{
  unsigned int msysprv[2];

  msysprv[0] = PRV$M_SYSPRV;
  msysprv[1] = 0;
  c$cks(sys$setprv(1,msysprv,0,0));
}

/*
 *  nntp_write
 *
 *  Synchronous write of a string to the net channel
 */

nntp_write(b)	char *b;
{
  static char obuf[NNTP_STRLEN+1];
  struct iosb write_iosb;
  int sts;

  strcpy(obuf,b);
  strcat(obuf,"\r\n");
  sts = sys$qiow(0,net_chan,IO$_WRITEVBLK,&write_iosb,0,0,obuf,strlen(obuf),0,(net_proto == CMUTCP),0,0);
  if (!(sts & 1) || (!(write_iosb.iostatus & 1))) {
    close_net();
    sprintf(client_msg,"NNTP: Lost connection to Server: %s",net_open_chan);
    net_read_status = 0;
    return(0);
    }
  return(1);
}

/*
 *  nntp_read
 *
 *  Timed read of the net channel
 */

jmp_buf nntp_env;

void cancel_nntp_read()
{
  close_net();
  sprintf(client_msg,"NNTP: Lost connection to Server: %s",net_open_chan);
  longjmp(nntp_env,1);
}

int nntp_read(b,timer) char *b; int timer;
{
  static char ibuf[X_BUF_SIZE + 1];
  struct iosb r_iosb;
  char *cp, *rp;
  int sts;

  *b = '\0';
  if (setjmp(nntp_env)) return(0);
  signal(SIGALRM,cancel_nntp_read);
  alarm(timer);

  if (decnetstream) sts = sys$qiow(0,net_chan,IO$_READVBLK | IO$M_MULTIPLE,&r_iosb,0,0,ibuf,X_BUF_SIZE,0,0,0,0);
  else sts = sys$qiow(0,net_chan,IO$_READVBLK,&r_iosb,0,0,ibuf,X_BUF_SIZE,0,0,0,0);

  alarm(0);
  if (!(sts & 1) || !(r_iosb.iostatus & 1) || !r_iosb.iosize) {
    close_net();
    sprintf(client_msg,"NNTP: Lost connection to Server: %s",net_open_chan);
    return(0);
    }
  ibuf[r_iosb.iosize] = '\0';
  rp = ibuf;
  do {
    if (cp = strchr(rp,'\r')) *cp++ = '\0';
    strcat(b,rp);
    rp = cp;
    } while(rp);
  return(1);
}

/*
 * nntp_read_line
 *
 * Read of 1 logical record from the net channel
 */

char lbuf[X_BUF_SIZE + 1] = {""};

int nntp_read_line(b,timer)	char *b; int timer;
{
  char *cp, *pp;

  *b = '\0';
  while (strlen(b) < 1024) {
    if (cp = strchr(lbuf,'\n')) {
      *cp = '\0';
      strcat(b,lbuf);
      strcat(b,"\n");
      ++cp;
      pp = lbuf;
      while (*pp++ = *cp++);
      return(1);
      }
    strcat(b,lbuf);
    if (!nntp_read(lbuf,timer)) return(0);
    }
  return(1);		/* glass@vixvax.mgi.com 6/3/91 */
}

/*
 *  wait_net_response
 *
 *  Wait for a response from remote system - cmd indicates that all text
 *  should be skipped until NNTP command response is obtained
 */

int wait_net_response(secs,cmd)	int secs,cmd;
{
  if (nntp_read_line(ibuf,secs)) {
    if (!cmd) return(1);
    if (sscanf(ibuf,"%d",&cmd_code) == 1) {
      if ((cmd_code == 400) || (cmd_code == 205)) {
	close_net();
	return(0);
	}
      else return(cmd_code);
      }
    else return(1);		/* glass@vixvax.mgi.com 6/3/91 */
    }
  else return(0);		/* glass@vixvax.mgi.com 6/3/91 */
}

#if UCX
static unsigned int get_local_adrs()
{
  static unsigned int localhost = 0;

  if (localhost) return(localhost);
  else {
    char *gotenv;
    short nam_chan, retlen;
    int comm = INETACP$C_TRANS * 256 + INETACP_FUNC$C_GETHOSTBYNAME;
    struct dsc$descriptor command = {4,DSC$K_CLASS_S, DSC$K_DTYPE_T,&comm};
    struct dsc$descriptor host_ad = {4,DSC$K_CLASS_S, DSC$K_DTYPE_T,&localhost};
    struct iosb nam_iosb;

    if (!(gotenv = getenv("UCX$INET_HOST"))) gotenv = "localhost";
    if (sys$assign(c$dsc("BG0:"),&nam_chan,0,0) & 1) {
      if (!(   (sys$qiow(0,nam_chan,IO$_ACPCONTROL,&nam_iosb,0,0,&command,c$dsc(gotenv),&retlen,&host_ad,0,0) & 1)
            && (nam_iosb.iostatus & 1)))
        localhost = (1 << 24) + 127;
      sys$dassgn(nam_chan);
      }
    return(localhost);
    }
  }

static unsigned int get_host_adrs(host_name)
  char *host_name;
{
  unsigned int hostadrs = 0;
  short nam_chan, retlen;
  int comm = INETACP$C_TRANS * 256 + INETACP_FUNC$C_GETHOSTBYNAME;
  struct dsc$descriptor command = {4,DSC$K_CLASS_S, DSC$K_DTYPE_T,&comm};
  struct dsc$descriptor host_ad = {4,DSC$K_CLASS_S, DSC$K_DTYPE_T,&hostadrs};
  struct iosb nam_iosb;

  if (sys$assign(c$dsc("BG0:"),&nam_chan,0,0) & 1) {
    if (!(   (sys$qiow(0,nam_chan,IO$_ACPCONTROL,&nam_iosb,0,0,&command,c$dsc(host_name),&retlen,&host_ad,0,0) & 1)
          && (nam_iosb.iostatus & 1))) hostadrs = 0;
      sys$dassgn(nam_chan);
    }
  else printf("Couldn't open BG0:\n");
  return(hostadrs);
}
#endif

/*
 *  open_net
 *
 *  Open network link
 */

int open_net(node,proto)	char *node; int proto;
{
  char netobj[128];

  strcpy(netobj,node);
  if (proto == CMUTCP) {
    if (sys$assign(c$dsc("IP:"),&net_chan,0,0) & 1) {
      if (   (sys$qiow(0,net_chan,IO$_CREATE,&read_iosb,0,0,netobj,119,0,1,0,300) & 1)
	  && (read_iosb.iostatus & 1))
	return(1);
      sys$dassgn(net_chan);
      }
    return(0);
    }

#if UCX
  else if (proto == UCXTCP) {
    if (sys$assign(c$dsc("BG:"),&net_chan,0,0) & 1) {
      sck_parm[0] = INET$C_TCP;
      sck_parm[1] = INET_PROTYP$C_STREAM;
      local_host.inet_family = INET$C_AF_INET;
      local_host.inet_port = 0;  /* dynamic port assignment */
      local_host.inet_adrs = get_local_adrs();
      lhst_adrs.lgth = sizeof local_host;
      lhst_adrs.hst = &local_host;
      if (   (sys$qiow(0,net_chan,IO$_SETMODE,&read_iosb,0,0,&sck_parm,0,&lhst_adrs,0,0,0) & 1)
          && (read_iosb.iostatus & 1)) {
        remote_host.inet_family = INET$C_AF_INET;
        remote_host.inet_port = htons(NNTP_PORT);
        if (!(remote_host.inet_adrs = get_host_adrs(netobj))) {
          char *c, *cd, resnum[4];
          unsigned int indx = 0, *i;

          c = netobj;
          while (c && (indx <= 3)) {
            if (cd = strchr(c,'.')) *cd = '/0';
            resnum[indx++] = atoi(c);
            if (cd) {
              *cd++ = '.';
              c = cd;
              }
            else c = 0;
            }
          if (c || indx != 4) {
            sys$qiow(0,net_chan, IO$_DEACCESS | IO$M_SHUTDOWN,read_iosb,0,0,0,0,0,UCX$C_DSC_ALL,0,0);
            sys$dassgn(net_chan);
            return(0);
            }
          i = &resnum[0];
          remote_host.inet_adrs = *i;
          }
        rhst_adrs.lgth = sizeof remote_host;
        rhst_adrs.hst = &remote_host;
        printf("Calling %08x\n",remote_host.inet_adrs);
        if (   (sys$qiow(0,net_chan,IO$_ACCESS,&read_iosb,0,0,0,0,&rhst_adrs,0,0,0) & 1)
            && (read_iosb.iostatus & 1)) return(1);
        printf("iosb.iostatus=%08x\n",read_iosb.iostatus);
        printf("Failed open_net\n");
        sys$qiow(0,net_chan, IO$_DEACCESS | IO$M_SHUTDOWN,read_iosb,0,0,0,0,0,UCX$C_DSC_ALL,0,0);
        }
      sys$dassgn(net_chan);
      }
    return(0);
    }
#endif

#if TWG || MULTINET || TCPWARE
  else if ((proto == WINTCP) || (proto == MULTINETTCP) || (proto == TCPWARETCP)) {
    if (   ((proto == WINTCP) && (sys$assign(c$dsc("INET:"),&net_chan,0,0) & 1))
	|| ((proto == MULTINETTCP) && (sys$assign(c$dsc("INET0:"),&net_chan,0,0) & 1))
	|| ((proto == TCPWARETCP) && (sys$assign(c$dsc("INET0:"),&net_chan,0,0) & 1))) {
      if (   (sys$qiow(0,net_chan,IO$_SOCKET,&read_iosb,0,0,
		    AF_INET,SOCK_STREAM,0,0,0,0) & 1)
	  && (read_iosb.iostatus & 1)) {

	if (!(dest_host = gethostbyname(netobj))) {
	  int h[4], i;

	  if (sscanf(netobj,"%d.%d.%d.%d",&h[0],&h[1],&h[2],&h[3]) != 4) {
	    if (debugging) fprintf(fpd,"open_net: gethostbyname returned null\n");
	    sys$dassgn(net_chan);
	    return(0);
	    }
	  for (i = 0 ; i < 4 ; ++i)
	    if (h[i] < 0 || h[i] > 255) {
	      if (debugging) fprintf(fpd,"Incorrect format of internet address\n");
	      return(0);
	      }
	  data_socket.sin_addr.s_addr = (h[3] << 24) + (h[2]  << 16) + (h[1] << 8) + h[0];
	  }
#if TWG
        else {
          addr_pointer = *((u_long *)(dest_host->h_addr)); /*kjb@calmasd.prime.com*/
          data_socket.sin_addr.s_addr = *(u_long *) addr_pointer; /*kjb@calmasd.prime.com*/
          }
#else
	else data_socket.sin_addr.s_addr = *((u_long *)(dest_host->h_addr));
#endif
	data_socket.sin_family = AF_INET;
	data_socket.sin_port = htons(NNTP_PORT);
	if (   (sys$qiow(0,net_chan,IO$_CONNECT,&read_iosb,0,0,
			 &data_socket,sizeof(data_socket),0,0,0,0) & 1)
	    && (read_iosb.iostatus & 1))
	  return(1);
	}
      }
    if (debugging) fprintf(fpd,"open_net: io$_connect or io$_Socket error\n");
    if (debugging)
      fprintf(fpd,"open_net: read_iosb.iostatus %x \n",read_iosb.iostatus);
    sys$dassgn(net_chan);
    return(0);
    }
#endif

  else {
    sprintf(netobj,"%s::\"%s\"",node,taskstr);
    return(sys$assign(c$dsc(netobj),&net_chan,0,0) & 1);
    }
}

/*
 *  close_net
 *
 *  Close network link
 */

void close_net()
{
  if (net_proto) {
    sys$cancel(net_chan);
    sys$qiow(0,net_chan,IO$_DELETE,0,0,0,0,0,0,0,0,0);
    }
  sys$dassgn(net_chan);
}

/*
 *  toggle_link
 *
 *  Write a non-command to the net chan and wait for the command-not-recognised
 *  response from the server.
 */

int toggle_link()
{
  if (!nntp_write("STAT") || !(wait_net_response(RESP_TIMER,1))) {
    if (*net_open_chan) {
      close_net();
      printf("Lost connection to NEWS SERVER ([%c] %s)",
		    (net_proto ? 'T' : 'D'),net_open_chan);
      *net_open_chan = 0;
      }
    return(0);
    }
  return(1);
}

/*
 *  lower_case
 *
 *  convert a string to lower case
 */

char *lower_case(s)	char *s;
{
  char *save = s;

  while (*s) {
    *s = tolower(*s);
    s++;
    }
  return(save);
}

/*
 *  open_rem_chan
 *
 *  Open channel to remote NNTP server on host name node, using protocol
 *  proto_num (6 == TCPWARETCP, 3 == MULTINETTCP, 2 == WINTCP,
 *  1 == CMUTCP, 0 = DECNET)
 */

int open_rem_chan(node,proto_num)	char *node; int proto_num;
{
  int reply;

  lower_case(node);
  if (*net_open_chan) {
    if ((proto_num != net_proto) || (strcmp(node,net_open_chan))) {
      close_net();
      net_read_status = 0;
      *net_open_chan = 0;
      }
/*
    else toggle_link();
*/
    }

  if (!*net_open_chan) {
    strcpy(net_open_chan,node);
    if (!open_net(net_open_chan,proto_num)) {
      net_read_status = 0;
      *net_open_chan = 0;
      if (debugging) fprintf(fpd,"Net: Could not open network channel\n");
      return(0);
      }
    else net_proto = proto_num;
    if (!(reply = wait_net_response(CLIENT_TIMER,1))) {
      close_net();
      net_read_status = 0;
      *net_open_chan = 0;
      if (debugging) fprintf(fpd,"Net: No reply received in time\n");
      return(0);
      }
    if (reply!=200) {
      printf("Postings not allowed to NEWS SERVER: %s",net_open_chan);
      if (reply==201) {
	printf(",only transfer allowed.\n");
	return(1);
	}
      printf("\n");
      return(0);
      }
    }
  return(1);
}

/*
 * premature_break
 *
 * Cleanup up following error or forced stop of the process.
 */

premature_break(type,fpr,fpi,fpw,logging,lines_written)	FILE *fpr, *fpi, *fpw; int lines_written,logging,type;
{
  char inpstr[132],
       type_text[3][43]={"Error: Connection lost",
			 "Information: Max. items offered",
			 "Information: Exit forced by system manager"};
  FILE *fpl;

  if (logging){
    dat();
    fpl = fopen(logfile,"a+");
    if(offered != 0) {
      fprintf(fpl,"%s %s: nntpxmit: %s %d offered, %d accepted, %d rejected, %d failed, %d expired",
		datum, lhost, node, offered, accepted, rejected, failed, expired);
		/**	Report on xfer rates: Earle Ake ake@dayton.saic.com */
      if (total_elapsed_time > 0)
        fprintf(fpl,", %d bytes, %.2f secs (%d bytes/sec)",
         total_bytes_sent,(float) total_elapsed_time/100.0,(total_bytes_sent*100)/total_elapsed_time);
      }
    fprintf(fpl,"%s %s: nntpxmit: %s, %s, transfer aborted.\n",datum, lhost, node,type_text[type]);
    fprintf(fpl,"\n");
    fclose(fpl);
    }
  if (!idx) {
    if (lines_written && fpw) {
      while (fgets(inpstr,132,fpr)) fprintf(fpw,"%s",inpstr);
      fclose(fpw);
      fclose(fpr);
      delete(idfile);
      rename(tmpf,idfile);
      }
    else {
      fclose(fpw);
      delete(tmpf);
      }
    }
  if (type == MAX_OFFERED) {
    int len = 3;
    char exec_mode = PSL$C_EXEC;
    unsigned int loglist[4]={(LNM$_STRING << 16) + 3,"YES",&len,0};

    sys$crelnm(0,c$dsc(LOG_TABLE),c$dsc("NNTP_MORE_TO_OFFER"),&exec_mode,&loglist);
    }
  if (debugging) fclose(fpd);
  if (idx) close_id_file();
  else fclose(fpr);
  fclose(fpi);
  exit(1);
}

/*
 *  util_dir
 *
 *  Convert a newsgroup name to a VMS directory string
 */

char dir_result[SUBJLEN + SUBJLEN];

char *util_dir(input)	char *input;
{
  char *p = dir_result,
       *in = input;

  while (*in) {
    if (isalnum(*in) || (*in == '-') || (*in == '.')) *p++ = *in++;
    else {
      *p++ = '_';
      if (*in == '_') *p++ = '_';
      else if (*in < '0') *p++ = (*in - '!') + 'A';
      else if (*in < 'A') *p++ = (*in - ':') + '0';
      else if (*in < 'a') *p++ = (*in - '[') + 'P';
      else *p++ = (*in - '{') + 'V';
      in++;
      }
    }
  *p = '\0';
  return(dir_result);
}

/*
 *  open_server_files
 *
 *  Open the item and newsgroup files for sharing
 */

int open_server_files()
{
  itmfab = cc$rms_fab;
  itmfab.fab$b_fac = FAB$M_GET;
  itmfab.fab$l_fna = ITM_FILENAME;
  itmfab.fab$b_fns = strlen(itmfab.fab$l_fna);
  itmfab.fab$b_shr = FAB$M_SHRDEL | FAB$M_SHRGET | FAB$M_SHRPUT | FAB$M_SHRUPD;

  itmrab = cc$rms_rab;
  itmrab.rab$l_fab = &itmfab;
  itmrab.rab$b_krf = 0;
  itmrab.rab$b_ksz = 8;
  itmrab.rab$l_ubf = &newsitm;
  itmrab.rab$w_usz = sizeof newsitm;
  itmrab.rab$l_rbf = &newsitm;
  itmrab.rab$w_rsz = sizeof newsitm;

  grpfab = cc$rms_fab;
  grpfab.fab$b_fac = FAB$M_GET;
  grpfab.fab$l_fna = GRP_FILENAME;
  grpfab.fab$b_fns = strlen(grpfab.fab$l_fna);
  grpfab.fab$b_shr = FAB$M_SHRDEL | FAB$M_SHRGET | FAB$M_SHRPUT | FAB$M_SHRUPD;

  grprab = cc$rms_rab;
  grprab.rab$l_fab = &grpfab;
  grprab.rab$l_ubf = &newsgrp;
  grprab.rab$w_usz = sizeof newsgrp;
  grprab.rab$l_rbf = &newsgrp;
  grprab.rab$w_rsz = sizeof newsgrp;

  if (	 !(sys$open(&itmfab) & 1)
      || !(sys$connect(&itmrab) & 1)
      || !(sys$open(&grpfab) & 1)
      || !(sys$connect(&grprab) & 1)) return(0);
  return(1);
}

/*
 *  main
 */

int  main(argc,argv)
  int argc; char *argv[];
{
  char proto[32], l_id[IDLEN + 4], inpline[MAXIDLEN], inpline0[512],
       qual[512], itm_fname[256], cmd[132], *cp;
  int  proto_num = 0, lines_written = 0, have_open_channel_OK = 0, logging = 0,
       reply, i, dbg_retval;
  unsigned int trnlist[4];
  unsigned short len, qual_len;
  FILE *fpr, *fpi, *fpl, *fpw;
  $DESCRIPTOR(qual_dsc,qual);

  *taskstr = '\0';
  *node = *proto = *idfile = *logfile = *dbgfile = '\0';
				/* interpret program arguments */
  if (argc < 2) {
    printf("Usage NNTP_XMIT [-s] [-x] [-t decnet-task] node proto idfile [log] [dbgfile]\n");
    printf("  -or-\n");
    printf("Usage: NNTP_XMIT/NODE=node/PROTOCOL=proto/IDFILE=idfile[/LOGFILE=log]\n");
    printf("            [/TASK=task][/DEBUGFILE=dbgfile][/INDEXID][/STREAM]\n");
    exit(1);
    }
  if (*argv[1] == '/') {
    char mcommand[1024];
    register int argindex;

    strcpy(mcommand,"INVOKENNTP");
    for (argindex = 1; argindex < argc; argindex++ )
    {
	strcat(mcommand," ");
	strcat(mcommand,argv[argindex]);
    }
    if (!(cli$dcl_parse(c$dsc(mcommand),&nntpxmitcmd,0,0,0) & 1)) exit(1);

    decnetstream = (cli$present(c$dsc("STREAM")) & 1);
    idx = (cli$present(c$dsc("INDEXID")) & 1);
    if (cli$get_value(c$dsc("NODE"),&qual_dsc,&qual_len) & 1) {
      qual[qual_len] = '\0';
      strcpy(node,qual);
      }
    if (cli$get_value(c$dsc("PROTOCOL"),&qual_dsc,&qual_len) & 1) {
      qual[qual_len] = '\0';
      strcpy(proto,qual);
      }
    if (cli$get_value(c$dsc("IDFILE"),&qual_dsc,&qual_len) & 1) {
      qual[qual_len] = '\0';
      strcpy(idfile,qual);
      }
    if (cli$get_value(c$dsc("LOGFILE"),&qual_dsc,&qual_len) & 1) {
      qual[qual_len] = '\0';
      strcpy(logfile,qual);
      logging = 1;
      }
    if (cli$get_value(c$dsc("DEBUGFILE"),&qual_dsc,&qual_len) & 1) {
      qual[qual_len] = '\0';
      strcpy(dbgfile,qual);
      debugging = 1;
      }
    if (cli$get_value(c$dsc("TASK"),&qual_dsc,&qual_len) & 1) {
      qual[qual_len] = '\0';
      if (qual_len) {
        if (!strchr(qual,'=')) strcpy(taskstr,"TASK=");
        else *taskstr = '\0';
        strcat(taskstr,qual);
	cp = taskstr;  --cp; while (*++cp) *cp = toupper(*cp);
        }
      }
    }
  else {
    for (i = 1 ; i < argc ; ++i) {
      if (*argv[i] == '-') {
        cp = argv[i];
        ++cp;
        if (tolower(*cp) == 'h') {
          printf("NNTP_XMIT\n");
          printf("Usage NNTP_XMIT [-s] [-x] [-t decnet-task] node proto idfile [log] [dbgfile]\n");
          printf("  -s - used with DECnet protocol to use stream Input Qio calls\n");
          printf("       (used to connect to Ultrix NNTP over DECnet)\n");
          printf("       default is ok for remote VMS systems\n");
          printf("\n");
          printf("  -x - used to read indexed message-id files rather than the default\n");
          printf("       sequential files\n");
          printf("       default is sequentail access routines\n");
          printf("\n");
          printf("  -t decnet-task   - the default task connect string is TASK=NNTP\n");
          printf("       other strings may be specified instead\n");
          printf("\n");
          printf("  node - name of node to connect to\n");
          printf("\n");
          printf("  proto - protocol to use \"tcp\" for tcp/ip, \"decnet\" for DECnet\n");
          printf("\n");
          printf("  idfile - name of message-id file to read\n");
          printf("\n");
          printf("  log - optional - name of log file to record activity of the run\n");
          printf("\n");
          printf("  dbgfile - optional - name of debug output file\n");
          printf("\n");
          printf("\n");
          printf("  e.g. NNTP_XMIT -x tcp.node.domain tcp tcpid.idx\n");
          printf("\n");
          printf("       NNTP_XMIT -s -x -t NNTP= ultrixnode decnet ultrix.idx\n");
          printf("\n");
          exit(1);
          }
        if (tolower(*cp) == 't') {
          if (++i < argc) {
            strcpy(taskstr,argv[i]);
            cp = taskstr;  --cp; while (*++cp) *cp = toupper(*cp);
            }
          else {
            *idfile = 0;
            break;
            }
          }
        else {
          while (*cp) {
            if (tolower(*cp) == 'x') idx = 1;
            if (tolower(*cp) == 's') decnetstream = 1;
            ++cp;
            }
          }
        }
      else if (!*node) strcpy(node,argv[i]);
      else if (!*proto) strcpy(proto,argv[i]);
      else if (!*idfile) strcpy(idfile,argv[i]);
      else if (!*logfile) {
        strcpy(logfile,argv[i]);
        logging = 1;
        }
      else if (!*dbgfile) {
        strcpy(dbgfile,argv[i]);
        debugging = 1;
        }
      else {
        *idfile = 0;
        break;
        }
      }
    }

  if (!*idfile) {
    printf("Usage: NNTP_XMIT [-s] [-x] [-t decnet-task] node proto idfile [log] [dbgfile]\n");
    exit(1);
    }

  if (!open_server_files()) {
    printf("FileError: Cannot open NEWS datafiles\n");
    exit(1);
    }

  if (	 (cp = getenv("NEWS_ADDRESS"))
#ifdef MULTINET
      || (cp = getenv("MULTINET_SMTP_HOST_NAME"))
      || (cp = getenv("MULTINET_HOST_NAME"))
#endif
#ifdef TCPWARE
      || (cp = getenv("SMTP_FROM_DOMAIN"))
      || (cp = getenv("TCPIP_DOMAINNAME"))
#endif
      || (cp = getenv("INTERNET_HOST_NAME")))
    lower_case(strcpy(lhost,cp));
  else {
    printf("NEWS_ADDRESS not defined - exiting\n");
    exit(1);
    }

#if CMU
  if (tolower(*proto) == 't') proto_num = CMUTCP;
  if (tolower(*proto) == 'c') proto_num = CMUTCP;
#endif
#if UCX
  if (tolower(*proto) == 't') proto_num = UCXTCP;
  if (tolower(*proto) == 'u') proto_num = UCXTCP;
#endif
#if TWG
  if (tolower(*proto) == 't') proto_num = WINTCP;
  if (tolower(*proto) == 'w') proto_num = WINTCP;
#endif
#if MULTINET
  if (tolower(*proto) == 't') proto_num = MULTINETTCP;
  if (tolower(*proto) == 'm') proto_num = MULTINETTCP;
  if (tolower(*proto) == 's') proto_num = MULTINETTCP;
#endif
#if TCPWARE
  if (tolower(*proto) == 't') proto_num = TCPWARETCP;
  if (tolower(*proto) == 'p') proto_num = TCPWARETCP;
#endif

  if (proto_num != DECNET) decnetstream = 0;
  else if (!*taskstr) {
    char logname[132],
	 *gotenv;

    sprintf(logname,"NEWS_%s_TASK",node);
    cp = logname;  --cp; while (*++cp) *cp = toupper(*cp);
    if (gotenv = getenv(logname)) strcpy(taskstr,gotenv);
    else strcpy(taskstr,"TASK=NNTP");
    cp = taskstr;  --cp; while (*++cp) *cp = toupper(*cp);
    }
  sysprv();
  if (	 (idx && !open_id_file(idfile))
      || (!idx && !(fpr=fopen(idfile,"r")))) {
    nosysprv();
    printf("FileError: Cannot open file: %s\n",idfile);
    exit(1);
    }

  if (!idx) {
    fgetname(fpr,idfile);
    strcpy(tmpf,idfile);
    if (cp = strchr(tmpf,';')) *cp = '\0';
    if (cp = strrchr(tmpf,']')) cp++;
    else if (cp = strrchr(tmpf,':')) cp++;
    else cp = tmpf;
    sprintf(cp,"NNTP_%X.TMP",getpid());
    fpw = fopen(tmpf,"w");
    }

  if (debugging) fpd = fopen(dbgfile,"w");

  printf("\tConnecting to NEWS SERVER on node %s ...\n",node);
  while (   (idx && read_id(inpline))
	 || (!idx && fgets(inpline,MAXIDLEN,fpr))) {
    if (cp = strchr(inpline,'\n')) *cp = '\0';

    if ((accepted + failed) > NNTP_MAX_OFFER) {
      nntp_write("QUIT");
      wait_net_response(CLIENT_TIMER,1);
      GO_DIE(MAX_OFFERED)
      }
    if (manager_say_stop()) {
      nntp_write("QUIT");
      wait_net_response(CLIENT_TIMER,1);
      GO_DIE(MANAGER_STOP)
      }

    c$cks(1);   /* fix memory leak RRS 21-Feb-1992 */
    for (i = 0; i < (IDLEN + 4); ++i) l_id[i] = '\0';
    strncpy(l_id,inpline,IDLEN);

    itmrab.rab$l_kbf = l_id;
    itmrab.rab$b_ksz = IDLEN + 4;
    itmrab.rab$b_krf = 1;
    itmrab.rab$l_rop = RAB$M_KGE | RAB$M_RRL | RAB$M_NLK;
    itmrab.rab$b_rac = RAB$C_KEY;

    grprab.rab$l_kbf = &(newsitm.itm_grp);
    grprab.rab$b_ksz = 4;
    grprab.rab$b_krf = 1;
    grprab.rab$l_rop = RAB$M_RRL | RAB$M_NLK;
    grprab.rab$b_rac = RAB$C_KEY;

    dbg_retval = sys$get(&itmrab);
    if (   !(dbg_retval & 1)
	|| strcmp(l_id,newsitm.itm_id)) {
      lower_case(l_id);
      dbg_retval = sys$get(&itmrab);
      if (   !(dbg_retval & 1)
	  || strcmp(l_id,newsitm.itm_id)) {
	if (debugging) fprintf(fpd,"ItemAccess: Cannot locate item %s\n",inpline);
	if (idx) delete_id();
	expired++;
	continue;
	}
      }
    if (!(sys$get(&grprab) & 1)) {
      if (debugging) fprintf(fpd,"GroupAccess: Cannot locate group %s\n",inpline);
      if (idx) delete_id();
      continue;
      expired++;
      }
    sprintf(itm_fname,Itm_template,util_dir(newsgrp.grp_name),newsitm.itm_num);
    if (!(fpi = fopen(itm_fname,"r"))) {
      if (debugging) {
	fprintf(fpd,"FileAccess: Cannot open item file %s\n",itm_fname);
	fprintf(fpd, "Retry: %s\n",inpline);
	}
      if (!idx) fprintf(fpw,"%s\n",inpline);
      ++lines_written;
      continue;
      }

    if (debugging && !have_open_channel_OK)
      fprintf(fpd,"X> Try open channel to remote node\n");
    if (!open_rem_chan(node,proto_num)) {
      if (debugging) fprintf(fpd,"Cannot connect to node, retry: %s\n",inpline);
      GO_DIE(ERROR)
      }

    sprintf(cmd,"IHAVE %s",inpline);
    have_open_channel_OK = 1;

    if (debugging) fprintf(fpd,"X> %s<",cmd);
    offered++;
    if (   !nntp_write(cmd)
	|| !(reply = wait_net_response(CLIENT_TIMER,1))) {
      if(debugging) fprintf(fpd, "No reply received, retry: %s\n",inpline);
      GO_DIE(ERROR)
      }
    if (debugging) fprintf(fpd,"%d>", reply);
    if ((reply == 435) || (reply == 437)) {
      if (debugging) fprintf(fpd," (Rejected: dont send");
      if (reply == 437) fprintf(fpd," and dont retry");
      rejected++;
      if (idx) delete_id();
      fclose(fpi);
      if (debugging) fprintf(fpd,")\n");
      continue;
      }
    if (reply != 335) {
      failed++;
      fclose(fpi);
      if (!idx) {
	fprintf(fpw,"%s\n",inpline);
	++lines_written;
	}
      if (debugging) fprintf(fpd,"\n");
      continue;
      }
    if (debugging) fprintf(fpd,"(send article)");
    bytes_sent = 0;
    elapsed_time(1);
    while (fgets(inpline0,512,fpi)) {
      char *t;
      if (*inpline0 == '.') memmove(inpline0 + 1,inpline0,strlen(inpline0) + 1);
      if (t = strchr(inpline0,'\n')) *t = '\0';
      if (!nntp_write(inpline0)) GO_DIE(ERROR)
      bytes_sent += strlen(inpline0);
      }
    fclose(fpi);
    if (!nntp_write(".") || (!(reply = wait_net_response(CLIENT_TIMER,1)))) GO_DIE(ERROR)
    total_bytes_sent += bytes_sent;
    total_elapsed_time += elapsed_time(0);
    if (debugging) fprintf(fpd," %d",reply);
    if (reply == 435  || reply == 437) {
      if (debugging) fprintf(fpd," (Rejected: dont send");
      if (reply == 437) fprintf(fpd," and dont retry");
      if (debugging) fprintf(fpd,"\n");
      rejected++;
      if (idx) delete_id();
      continue;
      }
    if (reply == 235) {
      accepted++;
      if (debugging) fprintf(fpd,"(accepted)\n");
      if (idx) delete_id();
      continue;
      }
    if (!idx) {
      fprintf(fpw,"%s\n",inpline);
      ++lines_written;
      }
    if (debugging) fprintf(fpd,"(failed)\n");
    failed++;
    }
  nntp_write("QUIT");
  wait_net_response(CLIENT_TIMER,1);
  if (idx) close_id_file();
  else {
    fclose(fpr);
    fclose(fpw);
    delete(idfile);
    if (lines_written) rename(tmpf,idfile);
    else delete(tmpf);
    }
  if (logging){
    dat();
    fpl = fopen(logfile,"a+");
    fprintf(fpl,"%s %s: nntpxmit: %s %d offered, %d accepted, %d rejected, %d failed, %d expired",
		datum, lhost, node, offered, accepted, rejected, failed, expired);
		/**	Report on xfer rates: Earle Ake ake@dayton.saic.com */
    if (total_elapsed_time > 0)
      fprintf(fpl,", %d bytes, %.2f secs (%d bytes/sec)",
       total_bytes_sent,(float) total_elapsed_time/100.0,(total_bytes_sent*100)/total_elapsed_time);
    fprintf(fpl,"\n");
    fclose(fpl);
    }
  if (debugging) fclose(fpd);
}
